Scriptname kuFOFQS extends Quest  

Function debugTrace (String debugMessage, Int debugLevel = 1, Bool notify = False)
	If debugMode && debugMode >= debugLevel
		If notify
			Debug.Notification ("!kuFOFQS." + debugMessage)
		EndIf
		Debug.Trace ("!kuFOFQS." + debugMessage)
	EndIf
EndFunction

Bool Property started Auto
Float version
Float lastVersion
Float lastOnUpdateTimeUpdate
Bool Property resetNow Auto
Int Property debugMode Auto
Int Property testCase Auto
Bool stoppedThenStarted
Float updateInterval
Bool Property uninstallNow Auto
Int menuMode
Bool inited
Message Property uninstallMessage Auto
MQ00QuestScript Property questMQ00 Auto

Event OnInit ()
	If !inited
		inited = True
		Int oldDebugMode = debugMode
		;show debugs in the log during initialisations. the current debugMode is saved.
		;and the debugMode set high, so that debugs are logged.
		debugMode = 10
		debugTrace ("OnInit")
		RegisterForSingleUpdate (1)
		;restore the previous debugMode.
		debugMode = oldDebugMode
	EndIf
EndEvent

Event OnMenuMode (Int id = 1)
	If menuMode != id
		menuMode = id
		debugTrace ("OnMenuMode " + menuMode)
	EndIf
EndEvent

Event OnUpdate ()
	;set version and update interval
	version = 0.65
	updateInterval = 0.25
	If questMQ00.GetStage () >= 5
		Float timeNow = Utility.GetCurrentRealTime ()
		If !menuMode && Math.abs (timeNow - lastOnUpdateTimeUpdate) > 60
			;if current number of seconds is > 1 minute from the last recorded second in the game, it's
			;likely that the game was reloaded.
			initGameLoaded ()

		ElseIf version != lastVersion || resetNow
			;there is a new version. or the player manually reset the mod by typing in the console:
			;SETPQV <modName> RESETNOW TRUE
			initGameLoaded ()

		ElseIf uninstallNow
			debugTrace ("OnUpdate uninstall")
			uninstallNow = False
			uninstall ()
			debugTrace ("uninstalled", 0, True)
			Stop ()

		Else
			;tick () is the main update function.
			;add your code in that function instead of here, to keep this starting block free of
			;mod-specific code.
			tick ()
			;register only for a single update.
			;if RegisterForUpdate () is used, and the mod is deactivated, the Event manager will keep
			;send OnUpdate () Event to the deactivated mod. problems will occur.
			;with RegisterForSingleUpdate (), if the mod is deactivated, the Event manager will fail
			;only once.
			RegisterForSingleUpdate (updateInterval)
		EndIf
		;save the current number of seconds since game-start.
		lastOnUpdateTimeUpdate = timeNow
		menuMode = 0

	Else
		RegisterForSingleUpdate (updateInterval)
	EndIf
EndEvent

Function initGameLoaded ()
	Int oldDebugMode = debugMode
	;show debugs in the log during initialisations.
	debugMode = 10
	Bool newReset
	If version != lastVersion || !started || resetNow
		If lastVersion && !stoppedThenStarted
			;if the mod was previously installed (i.e. lastVersion is set), then reset the mod
			;calling Stop () and Start ().
			;because aliases are only ever filled at OnInit () Event, this is needed in case the
			;updates adds or removes new aliases.
			debugTrace ("initGameLoaded reset from v" + lastVersion + " to v" + version)
			;do uninstall procedure
			uninstall (True)
			inited = False
			UnregisterForUpdate ()
			Stop ()
			Utility.Wait (1)
			;Start () will reinitiate the OnInit () Event.
			Start ()
			stoppedThenStarted = True
			newReset = True

		Else
			;this is a new mod activation or a player initiated reset.
			debugTrace ("initGameLoaded init v" + version)
			stoppedThenStarted = False
			lastVersion = version
			resetNow = False
			initNewVersion ()
			started = True
		EndIf
	EndIf

	If !newReset
		;if the game wasn't reset above (i.e. Stop () and Start () wasn't called), then this
		;must be a normal game load.
		debugTrace ("initGameLoaded loaded v" + version)
		initNewGameLoad ()
		;RegisterForSingleUpdate () needed here because
		RegisterForSingleUpdate (1)
	EndIf
	debugMode = oldDebugMode
EndFunction

Function initNewVersion ()
	If !started
		actors = new Actor [128]
		lastHealth = new Float [128]
		origConfidence = new Float [128]
		origAggression = new Float [128]
		origAssitance = new Float [128]
		lastInCombat = new Bool [128]
		highestDamage = new Float [128]
		timeYielded = new Float [128]
		userLastRiddenHorseFlees = True
		userPlayerTeammatesFlee = True
		userAlwaysYield = True;if False, npcs may not fight to the death. The chance is: health / base health.
		userYieldBeforeDamage = True;if False, actors will yield only after damage. If True, they will flee before combat.
		userFightToDeathChanceFactor = 1;if userAlwaysYield is False, this value can be used to modify the chance (i.e. health / base health) of them fioghting to the death.
	EndIf
EndFunction

Function initNewGameLoad ()
	playerRef = Game.GetPlayer ()
	timeQuestAliasesStopped = 0
	; debugTrace ("initNewGameLoad " + actors)
	; debugTrace ("initNewGameLoad " + origConfidence)
	; debugTrace ("initNewGameLoad " + origAggression)
	; debugTrace ("initNewGameLoad " + origAggression)
EndFunction

Function uninstall (Bool resetting = False)
	debugTrace ("uninstall")
	cleanAllActors ()
	removeAllFromFliers ()
	questAliases.Stop ()
	questFliers.Stop ()
	UnregisterForUpdate ()
	If !resetting
		If uninstallMessage
			uninstallMessage.Show ()
		Else
			Debug.Notification ("uninstalled")
		EndIf
	EndIf
EndFunction

Actor playerRef
Quest Property questAliases Auto
Float timeQuestAliasesStopped
Bool Property userLastRiddenHorseFlees Auto
Bool Property userPlayerTeammatesFlee Auto
Actor [] actors
Float [] lastHealth
Float [] origConfidence
Float [] origAggression
Float [] origAssitance
Bool [] lastInCombat
Float [] highestDamage
Float [] timeYielded
Armor Property armourYielded Auto
Armor Property armourYieldedToNPC Auto
Faction Property factionAutomaton Auto
Faction Property factionDragon Auto
VoiceType Property voiceTypeDraugr Auto
VoiceType Property voiceTypeSkeleton Auto
Faction Property factionCreature Auto
Int playerEnemyCount
Int playerEnemyCountMelee
Quest Property questFliers Auto
ObjectReference Property lastYieldMarker Auto
Bool noYield
Race Property raceWerewolf Auto
Bool noYieldCleaned
Race Property DragonRace Auto
Quest Property questPelagius Auto;da15. Doesn't require all to be dead but scenes look better when no one yields.
Bool Property userAlwaysYield Auto
Armor Property armourNeverYield Auto
Bool Property userYieldBeforeDamage Auto
Float Property userFightToDeathChanceFactor Auto
Quest Property questDA16 Auto;Waking Nightmare / Skull of corruption. Stage 50 requires all to be dead.

Function tick ()
	If questPelagius.GetStage () > 50 && questPelagius.GetStage () < 200
		Return
	EndIf
	If questDA16.GetStage () == 50
		debugTrace ("tick questDA16.GetStage () = 50", 2)
		LocationAlias questDA16Alias = questDA16.GetAlias (36) as LocationAlias
		Location questDA16Location = questDA16Alias.GetLocation ()
		If playerRef.IsInLocation (questDA16Location)
			Return
		EndIf
	EndIf

	noYield = playerRef.GetActorValue ("Health") / playerRef.GetBaseActorValue ("Health") < 0.25
	If noYield
		If !noYieldCleaned
			noYieldCleaned = True
			cleanAllActors (True)
			removeAllFromFliers ()
		EndIf

	Else
		noYieldCleaned = False
		If !questFliers.IsRunning ()
			questFliers.Start ()

		Else
			Float timeNow = Utility.GetCurrentRealTime ()
			If !questAliases.IsRunning ()
				debugTrace ("tick questAliases.Start ()")
				questAliases.Start ()
				timeQuestAliasesStopped = timeNow

			Else
				Float elapsed = timeNow - timeQuestAliasesStopped
				If elapsed < 0 || elapsed > 5
					debugTrace ("tick questAliases.Stop ()")
					timeQuestAliasesStopped = timeNow
					questAliases.Stop ()
				EndIf
			EndIf
		EndIf

		updatePlayerEnemyCount ()
		processActors ()
		cleanFliers ()
	EndIf
EndFunction

Event OnActorInit (Actor actorRef)
	If questMQ00.GetStage () >= 5 && actorRef && !noYield
		Bool valid = isActorValid (actorRef)
		debugTrace ("addActor " + actorRef + ") valid " + valid, 2)
		If valid
			Int i = findActorIndex (actorRef)
			If i == -1
				i = lastEmptyIndex
				If i > -1
					actors [i] = actorRef
					lastHealth [i] = actorRef.GetActorValue ("Health")
					origConfidence [i] = actorRef.GetActorValue ("Confidence")
					origAggression [i] = actorRef.GetActorValue ("Aggression")
					origAssitance [i] = actorRef.GetActorValue ("Assistance")
					lastInCombat [i] = actorRef.IsInCombat ()
					highestDamage [i] = 0
					timeYielded [i] = 0
					debugTrace ("addActor " + actorRef + ") index " + i + " with " + origConfidence [i] + " " + origAggression [i] + " " + origAssitance [i])
				EndIf
			EndIf
		EndIf
	EndIf
EndEvent

Bool Function isActorValid (Actor actorRef)
	Bool ok = False
	ActorBase actorBaseForm = actorRef.GetActorBase ()
	If !actorBaseForm.IsUnique () && !actorRef.IsDead ()
		Float distance = actorRef.GetDistance (playerRef)
		If distance < 10000
		If !actorRef.IsEssential () && !actorRef.IsGhost ()
		If !actorRef.IsInFaction (factionAutomaton) && !actorRef.IsInFaction (factionDragon)
		If actorRef.GetVoiceType () != voiceTypeDraugr && actorRef.GetVoiceType () != voiceTypeSkeleton && !actorRef.IsCommandedActor ()
		If actorRef.GetRace () != raceWerewolf && actorRef.GetRace () != DragonRace
		If userLastRiddenHorseFlees || actorRef != Game.GetPlayersLastRiddenHorse ()
		If userPlayerTeammatesFlee || !actorRef.IsPlayerTeammate ()
			ok = True
		EndIf
		EndIf
		EndIf
		EndIf
		EndIf
		EndIf
		EndIf
	EndIf
	Return ok
EndFunction

Function processActors ()
	Float timeNow = Utility.GetCurrentRealTime ()
	Int i = 0
	While i < actors.Length
		Actor actorRef = actors [i]
		If actorRef
			Float distance = actorRef.GetDistance (playerRef)
			Int itemCount = actorRef.GetItemCount (armourNeverYield)
			If !actorRef.IsDead () && distance < 10000 && itemCount < 1
				Float health = actorRef.GetActorValue ("Health")
				Bool inCombat = actorRef.IsInCombat ()
				lastInCombat [i] = inCombat
				itemCount = actorRef.GetItemCount (armourYielded)
				Actor combatTargetRef = actorRef.GetCombatTarget ()
				If (actorRef.IsPlayerTeammate () || actorRef.IsPlayersLastRiddenHorse ()) && (timeYielded [i] || itemCount)
					If timeNow > timeYielded [i] + 10 || (actorRef.GetDistance (playerRef) < 500 && actorRef.IsEssential ())
						debugTrace ("processActors remove yielded 'coz player teammate " + actorRef + ")")
						removeActor (actorRef, i, True)
						removeFromFliers (actorRef)
					EndIf

				ElseIf timeYielded [i] || itemCount
					; If health < lastHealth [i]
					; 	debugTrace ("processActors previously yielded" + actorRef + ") health dropped again")
					; 	;fight to death
					; 	removeActor (actorRef, i)
					; 	removeFromFliers (actorRef)

					; Else
					If (timeYielded [i] > 0 && timeNow > timeYielded [i] + 10) || (timeYielded [i] == -1 && inCombat) || (timeYielded [i] <= 0 && itemCount)
						debugTrace ("processActors previously yielded " + actorRef + ") inCombat " + inCombat)
						; actorRef.StopCombat ()
						; If combatTargetRef
						; 	combatTargetRef.StopCombat ()
						; EndIf
						calmActor (actorRef, i)
					EndIf

				ElseIf inCombat && combatTargetRef.GetRace () != raceWerewolf
					Float healthBase = actorRef.GetBaseActorValue ("Health")
					Float healthCombatTarget = healthBase
					Float healthBaseCombatTarget = healthCombatTarget
					Int itemCountCombatTarget = 0
					If combatTargetRef
						healthCombatTarget = combatTargetRef.GetActorValue ("Health")
						healthBaseCombatTarget = combatTargetRef.GetBaseActorValue ("Health")
						itemCountCombatTarget = combatTargetRef.GetItemCount (armourYielded) + combatTargetRef.GetItemCount (armourYieldedToNPC)
					EndIf
					If itemCountCombatTarget
						debugTrace ("processActors stopping combat 'coz target yielded already " + actorRef + ")")
						actorRef.StopCombat ()
						;actorRef.EvaluatePackage ()

					ElseIf userYieldBeforeDamage || health < healthBase
						Float damage = lastHealth [i] - health
						If damage > highestDamage [i]
							highestDamage [i] = damage
						EndIf
						If combatTargetRef && healthCombatTarget / healthBaseCombatTarget > 0.25
							If healthBase / healthCombatTarget < 0.25
								If combatTargetRef != playerRef
									distance = actorRef.GetDistance (combatTargetRef)
								EndIf
								Bool melee = isMelee (actorRef)
								If actorRef.IsInFaction (factionCreature) || (melee && (combatTargetRef != playerRef || playerEnemyCountMelee < 2)) || (!melee && distance < 750)
									debugTrace ("processActors yielded due to low-level" + actorRef + ")")
									; actorRef.StopCombat ()
									; If combatTargetRef
									; 	combatTargetRef.StopCombat ()
									; EndIf
									yieldActor (actorRef, i, health, healthBase)
								EndIf

							Else
								debugTrace ("processActors check for yield " + health + " < " + highestDamage [i] + " || " + health / healthBase + " < 0.25 || " + health / healthCombatTarget + " < 0.25" + actorRef + ")", 2)
								If health < highestDamage [i] || health / healthBase < 0.25 || (health / healthCombatTarget < 0.25 && healthBase / healthBaseCombatTarget > 0.25)
									debugTrace ("processActors yielded due to losing" + actorRef + ")")
									; If combatTargetRef
									; 	combatTargetRef.StopCombat ()
									; EndIf
									yieldActor (actorRef, i, health, healthBase)
								EndIf
							EndIf
						EndIf
					EndIf
				EndIf

				lastHealth [i] = health

			ElseIf actorRef.IsDead () || distance > 10000
				debugTrace ("processActors remove 'coz dead " + actorRef.IsDead () + " or " + distance + " > 10000 (" + actorRef + ")")
				removeActor (actorRef, i)
				removeFromFliers (actorRef)
			EndIf
		EndIf
		i += 1
	EndWhile
EndFunction

Function removeActor (Actor actorRef, Int actorIndex = -1, Bool removeTokens = False)
	Int i = actorIndex
	If i == -1
		i = findActorIndex (actorRef)
	EndIf
	If i > -1
		actorRef.SetActorValue ("Confidence", origConfidence [i])
		actorRef.SetActorValue ("Aggression", origAggression [i])
		actorRef.SetActorValue ("Assistance", origAssitance [i])
		debugTrace ("removeActor " + actorRef + ") with " + actorRef.GetActorValue ("Confidence") + " " + actorRef.GetActorValue ("Aggression") + " " + actorRef.GetActorValue ("Assistance"))
		If removeTokens
			Int itemCount = actorRef.GetItemCount (armourYielded)
			If itemCount
				actorRef.RemoveItem (armourYielded, itemCount)
			EndIf
			itemCount = actorRef.GetItemCount (armourYieldedToNPC)
			If itemCount
				actorRef.RemoveItem (armourYieldedToNPC, itemCount)
			EndIf
		EndIf
		actors [i] = None
		lastHealth [i] = 0
		origConfidence [i] = 0
		origAggression [i] = 0
		origAssitance [i] = 0
		lastInCombat [i] = 0
		highestDamage [i] = 0
		timeYielded [i] = 0
	EndIf
EndFunction

Function yieldActor (Actor actorRef, Int actorIndex = -1, Float health = 1.0, Float healthBase = 1.0)
	Bool yieldNow = userAlwaysYield
	Int itemCount
	;if this function is called before actor is damage, that means userYieldBeforeDamage is true, and the actor will yield.
	If !yieldNow && health < healthBase && !actorRef.IsPlayerTeammate () && !actorRef.IsPlayersLastRiddenHorse () && !actorRef.IsInFaction (factionCreature)
		;the chance of an actor fighting to the death only occurs after they are damaged.
		Float chance = health / healthBase * userFightToDeathChanceFactor
		Float random = Utility.RandomFloat ()
		If random > chance
			yieldNow = True

		Else
			;they will fight to the death
			debugTrace ("yieldActor " + actorRef + ") will never yield")
			itemCount = actorRef.GetItemCount (armourNeverYield)
			If itemCount < 1
				actorRef.AddItem (armourNeverYield, 1)
			EndIf
			Return
		EndIf
	EndIf
	Int i = actorIndex
	If i == -1
		i = findActorIndex (actorRef)
	EndIf
	If i > -1
		debugTrace ("yieldActor " + actorRef + ")", 1, True)
		Actor combatTargetRef = actorRef.GetCombatTarget ()
		If combatTargetRef
			combatTargetRef.StopCombat ()
			actorRef.StopCombatAlarm ()
		EndIf

		itemCount = actorRef.GetItemCount (armourYielded)
		If !itemCount
			actorRef.AddItem (armourYielded, 1)
		EndIf

		actorRef.SetActorValue ("Confidence", 0)
		actorRef.SetActorValue ("Aggression", 0)
		actorRef.SetActorValue ("Assistance", 0)
		timeYielded [i] = Utility.GetCurrentRealTime ()
	EndIf
EndFunction

Function calmActor (Actor actorRef, Int actorIndex = -1)
	Int i = actorIndex
	If i > -1
		Actor closestActor = Game.FindClosestActorFromRef (actorRef, 500)
		Bool attackAgain
		If closestActor
			Int relationshipRank = actorRef.GetRelationshipRank (closestActor)
			If relationshipRank > 0
				attackAgain = True
			EndIf
		EndIf
		If attackAgain || actorRef.GetItemCount (armourYielded) < 1
			debugTrace ("calmActor attackAgain " + actorRef)
			removeActor (actorRef, actorIndex, True)
			removeFromFliers (actorRef)

		Else
			debugTrace ("calmActor " + actorRef + ")", 1, True)
			Actor combatTargetRef = actorRef.GetCombatTarget ()
			If combatTargetRef
				combatTargetRef.StopCombat ()
				actorRef.StopCombatAlarm ()
			EndIf
			If timeYielded [i] != -1
				debugTrace ("calmActor " + actorRef)
				If actorRef.IsInFaction (factionCreature)
					actorRef.SetActorValue ("Confidence", 0)

				Else
					actorRef.SetActorValue ("Confidence", 1)
				EndIf
				actorRef.SetActorValue ("Aggression", 0)
				actorRef.SetActorValue ("Assistance", 0)
				actorRef.StopCombat ()
				; actorRef.EvaluatePackage ()
				If actorRef.GetItemCount (armourYielded)
					actorRef.EquipItem (armourYielded)

				ElseIf actorRef.GetItemCount (armourYieldedToNPC)
					actorRef.EquipItem (armourYieldedToNPC)
				EndIf
				timeYielded [i] = -1
				; questFliers.Stop ()
				; questFliers.Start ()
				; removeActor (actorRef, i)
				addActorToFliers (actorRef)
			EndIf
		EndIf
	EndIf
EndFunction

Function addActorToFliers (Actor actorRef)
	ReferenceAlias actorAlias = findActorAlias (actorRef)
	If actorAlias
		Actor aliasRef = actorAlias.GetActorReference ()
		If aliasRef != actorRef
			actorAlias.ForceRefTo (actorRef)
			;actorRef.AllowBleedoutDialogue (True)
		EndIf
	EndIf
EndFunction

ReferenceAlias Function findActorAlias (Actor actorRef)
	Int i = 0
	ReferenceAlias actorAlias
	Actor aliasRef
	ReferenceAlias emptyAlias
	Bool exitLoop
	Bool found
	While i < 20 && !exitLoop
		actorAlias = questFliers.GetAlias (i) as ReferenceAlias
		aliasRef = actorAlias.GetActorReference ()
		If aliasRef == actorRef
			exitLoop = True
			found = True

		ElseIf !aliasRef && !emptyAlias
			emptyAlias = actorAlias
		EndIf
		i += 1
	EndWhile
	If !found
		actorAlias = emptyAlias
	EndIf
	Return actorAlias
EndFunction

Function removeAllFromFliers ()
	Int i = 0
	ReferenceAlias actorAlias
	While i < 20
		actorAlias = questFliers.GetAlias (i) as ReferenceAlias
		actorAlias.Clear ()
		i += 1
	EndWhile
EndFunction

Function cleanFliers ()
	debugTrace ("cleanFliers")
	Int i = 0
	ReferenceAlias actorAlias
	Actor actorRef
	While i < 20
		actorAlias = questFliers.GetAlias (i) as ReferenceAlias
		actorRef = actorAlias.GetActorReference ()
		If actorRef
			If actorRef.IsDead () || actorRef.GetDistance (playerRef) > 10000
				debugTrace ("cleanFliers " + actorRef)
				actorAlias.Clear ()
			EndIf

		Else
			;actorAlias.Clear ()
		EndIf
		i += 1
	EndWhile
EndFunction

Function removeFromFliers (Actor actorRef)
	ReferenceAlias actorAlias = findActorAlias (actorRef)
	If actorAlias
		actorAlias.Clear ()
	EndIf
EndFunction

Function cleanAllActors (Bool removeTokens = False)
	debugTrace ("cleanAllActors")
	Int i = 0
	Bool cleaned
	While i < actors.Length
		Actor actorRef = actors [i]
		If actorRef
			removeActor (actorRef, i, removeTokens)
			cleaned = True
		EndIf
		i += 1
	EndWhile
	; If cleaned
	; 	questFliers.Stop ()
	; 	questFliers.Start ()
	; EndIf
EndFunction

Function lootFromDialogue (Actor actorRef)
	actorRef.OpenInventory (True)
EndFunction

Int lastEmptyIndex
Int Function findActorIndex (Actor actorRef, Bool returnEmpty = False)
	Int found = -1
	Int empty = -1
	Int i = 0
	While i < actors.Length && found == -1
		Actor actorRef2 = actors [i]
		If actorRef2 == actorRef
			found = i

		ElseIf !actorRef2 && empty == -1
			empty = i
			lastEmptyIndex = i
		EndIf
		i += 1
	EndWhile
	If found == -1 && returnEmpty
		found = empty
	EndIf
	Return found
EndFunction

Function updatePlayerEnemyCount ()
	playerEnemyCount = 0
	playerEnemyCountMelee = 0
	Int i = 0
	While i < actors.Length
		Actor actorRef = actors [i]
		If actorRef && actorRef.IsInCombat () && actorRef.GetActorValue ("Confidence") > 0 && actorRef.GetCombatTarget () == playerRef
			Float health = actorRef.GetActorValue ("Health")
			Float healthPlayer = playerRef.GetActorValue ("Health")
			If health / healthPlayer > 0.25
				playerEnemyCount += 1
				If isMelee (actorRef)
					playerEnemyCountMelee += 1
				EndIf
			EndIf
		EndIf
		i += 1
	EndWhile
EndFunction

Bool Function isMelee (Actor actorRef)
	Bool melee
	Int equippedItemTypeLeft = actorRef.GetEquippedItemType (0)
	Int equippedItemTypeRight = actorRef.GetEquippedItemType (1)
	If equippedItemTypeLeft < 7 || equippedItemTypeRight < 7
		melee = True
	EndIf
	Return melee
EndFunction
